import os
import re
import sys
from tempfile import gettempdir
from uuid import uuid4
import tomli


def test_version():
    with open("pyproject.toml", "rb") as f:
        pyproject_toml = tomli.load(f)
    with open("Cargo.toml", "rb") as f:
        cargo_toml = tomli.load(f)
    assert cargo_toml["package"]["version"] == pyproject_toml["project"]["version"]


def test_readme():
    files_under_test = ["../README.md", "README-pypi.md"]
    code_block_start = "```python"
    code_block_end = "```"
    code_block_regex = re.compile(
        f"{code_block_start}(.*?){code_block_end}", flags=re.DOTALL
    )
    for filename in files_under_test:
        with open(filename, "r") as f:
            content = f.read()
        codeblocks = code_block_regex.findall(content)
        codeblocks = "\n".join(codeblocks)
        tmpfile = f"{gettempdir()}/{uuid4().hex}.py"
        exit_code = None
        try:
            with open(tmpfile, "w") as f:
                f.write(codeblocks)
            exit_code = os.system(f"{sys.executable} {tmpfile}")
        finally:
            os.remove(tmpfile)
        assert exit_code == 0


if __name__ == "__main__":
    test_readme()



from typing import List, NamedTuple, Tuple, Union
import numpy as np
import pandas as pd
from .rormula import parse_wilkinson, eval_wilkinson, parse_arithmetic, eval_arithmetic


class SeparatedData(NamedTuple):
    numerical_cols: List[str]
    numerical_data: np.ndarray
    categorical_cols: List[str]
    categorical_data: np.ndarray


def separate_num_cat(
    data: pd.DataFrame,
) -> SeparatedData:
    """Separates numerical and categorical data for faster evaluation."""
    numerical = data.select_dtypes(include="number")
    categorical = data.select_dtypes(exclude="number")
    cat_cols = categorical.columns.to_list()
    if categorical.shape[1] == 0:
        categorical = categorical.to_numpy().astype("O")
    else:
        categorical = categorical.to_numpy()
    return SeparatedData(
        numerical.columns.to_list(),
        numerical.to_numpy(),
        cat_cols,
        categorical,
    )


class Wilkinson:
    def __init__(self, formula: str):
        self.ror = parse_wilkinson(formula)

    def eval(
        self, data: Union[pd.DataFrame, SeparatedData], skip_names: bool = False
    ) -> Tuple[List[str], np.ndarray]:
        if isinstance(data, SeparatedData):
            numerical_cols, numerical_data, categorical_cols, categorical_data = data
        else:
            (
                numerical_cols,
                numerical_data,
                categorical_cols,
                categorical_data,
            ) = separate_num_cat(data)

        names, resulting_data = eval_wilkinson(
            self.ror,
            numerical_data,
            numerical_cols,
            categorical_data,
            categorical_cols,
            skip_names=skip_names,
        )
        if names is None:
            names = []
        return names, resulting_data

    def eval_asdf(
        self, data: Union[pd.DataFrame, SeparatedData], skip_names: bool = False
    ):
        names, resulting_data = self.eval(data, skip_names=skip_names)
        return pd.DataFrame(data=resulting_data, columns=names)


class Arithmetic:
    def __init__(self, formula: str, name: str):
        self.ror = parse_arithmetic(formula)
        self.name = name

    def eval(self, data: pd.DataFrame) -> np.ndarray:
        numerical_cols = data.columns.to_list()
        numerical_data = data.to_numpy()

        resulting_data = eval_arithmetic(
            self.ror,
            numerical_data,
            numerical_cols,
        )

        return resulting_data

    def eval_asdf(self, data: pd.DataFrame):
        resulting_data = self.eval(data)
        if resulting_data.shape[0] == data.shape[0]:
            data[self.name] = resulting_data
        else:
            data = pd.DataFrame(data=resulting_data, columns=[self.name])
        return data



from functools import partial
from time import perf_counter
from typing import cast
import formulaic
from rormula import Wilkinson, SeparatedData
import rormula as ror
import numpy as np
import pandas as pd


FORMULA_STR_NUMERICAL = "a+b+a:b+c+d+c:d+e+f+e:f"
COLS_NUMERICAL = ["a", "b", "c", "d", "e", "f"]


def get_numerical_data(n_rows):
    return np.random.random((n_rows, len(COLS_NUMERICAL)))


def test_num_cat():
    n_rows = 100
    formula_str = f"{FORMULA_STR_NUMERICAL}+animal"
    animal_row = np.array(["dog", "cat", "horse", "okapi"] * (n_rows // 4))
    cols = COLS_NUMERICAL + ["animal"]
    data = pd.DataFrame(data=get_numerical_data(n_rows), columns=COLS_NUMERICAL)
    data[cols[-1]] = animal_row
    timing_and_test(data, formula_str)


def test_numerical():
    data_numerical = pd.DataFrame(data=get_numerical_data(100), columns=COLS_NUMERICAL)
    timing_and_test(data_numerical, FORMULA_STR_NUMERICAL)


def test_small_numerical():
    matrix = np.arange(6, dtype=np.float64).reshape((2, 3))
    data = pd.DataFrame(data=matrix, columns=["a", "b", "c"])
    rormula = Wilkinson("a+b+c")
    _, res = rormula.eval(data)
    ref = pd.concat(
        [pd.DataFrame(data=np.ones((2, 1)), columns=["Intercept"]), data], axis=1
    )
    np.allclose(res, ref)
    rormula = Wilkinson("(a+b):(a+c)")
    res = rormula.eval_asdf(data)
    ref_ = pd.concat(
        [
            pd.Series(data=[1, 1], name="Intercept"),
            pd.Series(data=[0, 9], name="a:a"),
            pd.Series(data=[0, 12], name="b:a"),
            pd.Series(data=[0, 15], name="a:c"),
            pd.Series(data=[2, 20], name="b:c"),
        ],
        axis=1,
    )
    np.allclose(res, ref_)
    rormula = Wilkinson("a+b+c+a:b+c^2")
    res = rormula.eval_asdf(data)
    ref = pd.concat(
        [
            ref,
            pd.Series(data=[0, 12], name="a:b"),
            pd.Series(data=[9, 36], name="c^2"),
        ],
        axis=1,
    )
    np.allclose(res, ref)


def test_missing_name_in_str():
    formular_str = FORMULA_STR_NUMERICAL
    cols = COLS_NUMERICAL + ["acdc"]
    n_rows = 6
    data = np.concatenate([get_numerical_data(n_rows), np.ones((n_rows, 1))], axis=1)
    data = pd.DataFrame(data=data, columns=cols)
    rormula = Wilkinson(formular_str)
    _, M_r = rormula.eval(data)
    assert np.allclose(data.to_numpy()[:, 1], M_r[:, 2])


def test_missing_name_in_col():
    formular_str = FORMULA_STR_NUMERICAL
    cols = COLS_NUMERICAL[:-1]
    data = get_numerical_data(6)[:, :-1]
    data = pd.DataFrame(data=data, columns=cols)
    rormula = Wilkinson(formular_str)
    try:
        rormula.eval(data)
        assert False
    except ValueError:
        pass
    except Exception:
        assert False


def timing(f, name, n_warmups=10, n_runs=100):
    res = None
    for _ in range(n_warmups):
        res = f()
    t = perf_counter()
    for _ in range(n_runs):
        res = f()
    t1 = perf_counter()
    print(f"{name} took {t1 - t:0.4f}s")
    return res


def timing_and_test(data, formula_str):
    rormula = Wilkinson(formula_str)
    # keeping data numerical and categorical data separated is faster
    separated_data = ror.separate_num_cat(data)
    M_r = timing(partial(rormula.eval, data=separated_data), "Rormula")
    M_r_asdf = timing(partial(rormula.eval_asdf, data=data), "Rormula asdf")

    assert M_r is not None
    assert M_r_asdf is not None
    names, M_r = M_r
    if len(names) == 0:
        return
    M_r = pd.DataFrame(data=M_r, columns=names)
    assert np.allclose(M_r, M_r_asdf)
    formula = formulaic.Formula(formula_str.replace("^", "**"))
    M_f = timing(partial(formula.get_model_matrix, data=data), "Formulaic")

    assert M_f is not None
    assert np.allclose(cast(pd.Series, M_f["e:f"]), M_r["e:f"])
    assert np.allclose(cast(pd.Series, M_f["f"]), M_r["f"])
    assert np.allclose(cast(pd.Series, M_f["c:d"]), M_r["c:d"])
    if "animal_dog" in M_r.columns:
        assert np.allclose(cast(pd.Series, M_f["animal[T.dog]"]), M_r["animal_dog"])
    assert np.allclose(cast(pd.Series, M_f["Intercept"]), M_r["Intercept"])


def test_more_formulas():
    def test(formula_str, extract_reference, extract_result):
        cols = ["alpha", "beta", "gamma", "eta", "theta", "omega"]
        data = np.random.random((100, len(cols)))
        data = pd.DataFrame(data=data, columns=cols)
        rormula = Wilkinson(formula_str)
        M_r = rormula.eval_asdf(data)
        a = extract_reference(data)
        b = extract_result(M_r)
        np.allclose(a, b)

    formula_str = "(beta):gamma"
    extract_reference = lambda data: data["beta"] * data["gamma"]
    extract_result = lambda M_r: M_r["beta:gamma"]
    test(formula_str, extract_reference, extract_result)

    formula_str = "(alpha + beta):gamma"
    extract_reference = lambda data: pd.concat(
        [data["alpha"] * data["gamma"], data["beta"] * data["gamma"]], axis=1
    )
    extract_result = lambda M_r: M_r[["alpha:gamma", "beta:gamma"]]
    test(formula_str, extract_reference, extract_result)

    formula_str = "(alpha + beta):(gamma + eta) + theta + omega"
    extract_reference = lambda data: pd.concat(
        [
            data["alpha"] * data["gamma"],
            data["beta"] * data["gamma"],
            data["alpha"] * data["eta"],
            data["beta"] * data["eta"],
            data["theta"],
            data["omega"],
        ],
        axis=1,
    )
    extract_result = lambda M_r: M_r[
        ["alpha:gamma", "beta:gamma", "alpha:eta", "beta:eta", "theta", "omega"]
    ]
    test(formula_str, extract_reference, extract_result)


def test_separated():
    data = pd.DataFrame(
        data=np.random.random((100, 3)),
        columns=["alpha", "beta", "gamma"],
    )
    separated_data = SeparatedData(
        numerical_cols=data.columns.to_list(),
        numerical_data=data.to_numpy(),
        categorical_cols=[],
        categorical_data=np.zeros((100, 0), dtype="O"),
    )
    rormula = Wilkinson("alpha + beta + alpha:gamma")
    names, mm = rormula.eval(separated_data)
    assert names == ["Intercept", "alpha", "beta", "alpha:gamma"]
    assert mm.shape == (100, 4)


if __name__ == "__main__":
    print("- test just numerical")
    test_numerical()
    print("- test numerical and categorical")
    test_num_cat()



from typing import List, NamedTuple, Optional, Sequence, Tuple
import numpy as np

class Wilkinson:
    pass

class SeparatedData(NamedTuple):
    numerical_cols: List[str]
    numerical_data: np.ndarray
    categorical_cols: List[str]
    categorical_data: np.ndarray

def parse_wilkinson(s: str) -> Wilkinson: ...
def eval_wilkinson(
    ror: Wilkinson,
    numerical_data: np.ndarray,
    numerical_cols: Sequence[str],
    cat_data: np.ndarray,
    cat_cols: Sequence[str],
    skip_names: bool = False,
) -> Tuple[Optional[List[str]], np.ndarray]: ...

class Arithmetic:
    pass

def parse_arithmetic(s: str) -> Arithmetic: ...
def eval_arithmetic(
    ror: Arithmetic,
    numerical_data: np.ndarray,
    numerical_cols: Sequence[str],
) -> np.ndarray: ...



from time import perf_counter
import numpy as np
import pandas as pd

from rormula import Arithmetic


def timing(f, name):
    t = perf_counter()
    res = None
    for _ in range(100):
        res = f()
    t1 = perf_counter()
    print(f"{name} took {t1 - t:0.4f}s")
    return res


def test_arithmetic():
    df = pd.DataFrame(
        data=np.random.random((100, 3)), columns=["alpha", "beta", "gamma"]
    )
    s = "beta*alpha - 1 + 2^beta + alpha / gamma"
    rormula = Arithmetic(s, "s")
    df_ror = rormula.eval_asdf(df.copy())
    pd_s = f's={s.replace("^", "**")}'
    assert df_ror.shape == (100, 4)
    assert np.allclose(df_ror, df.eval(pd_s))

    def eval_asdf():
        df_ror = df.copy()
        rormula.eval_asdf(df_ror)

    timing(eval_asdf, "eval_asdf")
    timing(lambda: df.eval(pd_s), "eval_pd")
    timing(lambda: rormula.eval(df), "eval_asdf")
    timing(lambda: df.eval(pd_s.split("=")[1]), "eval_pd")

    data = np.ones((100, 3))
    data[5, :] = 2.5
    data[7, :] = 2.5
    df = pd.DataFrame(data=data, columns=["alpha", "beta", "gamma"])
    s = "beta|alpha==2.5"
    rormula = Arithmetic(s, "reduced")
    res = rormula.eval_asdf(df)
    assert res.shape == (2, 1)
    assert np.allclose(res, 2.5)
    s = "beta|alpha>=2.5"
    rormula = Arithmetic(s, s)
    res = rormula.eval_asdf(df)
    assert res.shape == (2, 1)
    assert np.allclose(res, 2.5)
    s = "beta|alpha<2.5"
    rormula = Arithmetic(s, s)
    res = rormula.eval_asdf(df)
    assert res.shape == (98, 1)

    s = "((first_var|{second.var}==5.0) - (first_var|{second.var}==2.5)) / 4.0"
    data[7, :] = 5.0
    df = pd.DataFrame(data=data[:10, :2], columns=["first_var", "second.var"])
    rormula = Arithmetic(s, "reduced")
    res = rormula.eval_asdf(df)
    assert np.allclose(res.to_numpy().item(), (5.0 - 2.5) / 4.0)


def test_scalar_scalar():
    name = "test_scalar"
    data = np.random.random((100, 6)) * 1000
    df = pd.DataFrame(
        data=data, columns=["alpha", "beta", "gamma", "delta", "epsilon", "phi"]
    )
    s = "5/3 * alpha / beta * (0.2 / 200.0 / (29.22+gamma+epsilon+phi) / 1000)"
    rormula = Arithmetic(s, name)
    res = rormula.eval_asdf(df)
    ref = df.eval(s)
    np.allclose(res[name].to_numpy(), ref.values)
    s = "5/3"
    rormula = Arithmetic(s, name)
    res = rormula.eval_asdf(df)
    ref = df.eval(s)
    np.allclose(res[name].to_numpy(), ref)


if __name__ == "__main__":
    test_arithmetic()
    test_scalar_scalar()
